/*
 * ConversionHandler.cpp
 *
 *  Created on: 23. 2. 2014
 *      Author: tomas
 */

#include "ConversionHandler.h"

using namespace alib;
using namespace automaton;
using namespace grammar;
using namespace regexp;
using namespace sax;
using namespace std;

namespace conversions
{

ConversionHandler::ConversionHandler( list<Token> & tokens, const string & target, const string & algorithm ) : m_tokens( tokens )
{
    m_source = parseFormalismFromString( m_tokens.front( ).getData( ) );
    m_target = parseFormalismFromString( target );

    if( m_source == m_target )
        throw AlibException( "Cannot convert between same formalisms." );

    m_algorithm = parseAlgorithmFromString( algorithm );
    if( m_algorithm == DEFAULT )
        m_algorithm = getDefaultAlgorithm( );

    if( ! checkAlgorithm( ) )
        throw AlibException( "Invalid conversion. See help." );
}

void ConversionHandler::convert( ostream & out )
{
    if( m_source == FINITE_AUTOMATA )
        convertAutomata( out );
    if( m_source == REGULAR_EXPRESSION )
        convertRegExp( out );
    if( m_source == REGULAR_GRAMMAR )
        convertGrammar( out );
}

void ConversionHandler::convertGrammar( ostream & out )
{
    const UnknownGrammar ug = GrammarParser::parse( m_tokens );

    if( m_target == FINITE_AUTOMATA )
    {
        AbstractRGtoFAConverter* conv;
        RightRegularGrammar rrg;
        LeftRegularGrammar lrg;
        bool rightGrammar = true;

        try
        {
            GrammarFactory::buildRightRegularGrammar( ug );
        }
        catch( AlibException & e )
        {
            rightGrammar = false;
        }

        if( rightGrammar )
        {
            rrg = GrammarFactory::buildRightRegularGrammar( ug );
            conv = new RRGtoFAConverter( rrg );
        }
        else
        {
            lrg = GrammarFactory::buildLeftRegularGrammar( ug );
            conv = new LRGtoFAConverter( lrg );
        }

        conv->convert( ).toXML( out );
        delete conv;

    }
    else if( m_target == REGULAR_EXPRESSION )
    {
        AbstractRGtoREConverter* conv;
        RightRegularGrammar rrg;
        LeftRegularGrammar lrg;
        bool rightGrammar = true;

        try
        {
            GrammarFactory::buildRightRegularGrammar( ug );
        }
        catch( AlibException & e )
        {
            rightGrammar = false;
        }

        if( m_algorithm == BRZOZOWSKI_ALGEBRAIC && rightGrammar )
        {
            rrg = GrammarFactory::buildRightRegularGrammar( ug );
            conv = new RRGAlgebraic( rrg );
        }
        else if( m_algorithm == BRZOZOWSKI_ALGEBRAIC && ! rightGrammar )
        {
            lrg = GrammarFactory::buildLeftRegularGrammar( ug );
            conv = new LRGAlgebraic( lrg );
        }
        else throw AlibException( "wtf?" );

        conv->convert( ).toXML( out );
        delete conv;
    }
}
void ConversionHandler::convertRegExp( ostream & out )
{
    const RegExp regexp = RegExpParser::parse( m_tokens );

    if( m_target == FINITE_AUTOMATA )
    {
        AbstractREtoFAConverter* conv;
        if( m_algorithm == THOMPSON )
            conv = new Thompson( regexp );
        else if( m_algorithm == BRZOZOWSKI_DERIVATION )
            conv = new Brzozowski( regexp );
        else
            conv = new Glushkov( regexp );

        conv->convert( ).toXML( out );
        delete conv;

    }
    else if( m_target == REGULAR_GRAMMAR )
    {
        if( m_algorithm == BRZOZOWSKI_RRG )
        {
            AbstractREtoRRGConverter* conv;

            conv = new BrzozowskiDerivationRRG( regexp );
            conv->convert( ).toXML( out );
            delete conv;
        }
    }
}

void ConversionHandler::convertAutomata( ostream & out )
{
    const FSM fsm = AutomatonFactory::buildFSM( AutomatonParser::parse( m_tokens ) );

    if( m_target == REGULAR_EXPRESSION )
    {
        AbstractFAtoREConverter* conv;

        if( m_algorithm == BRZOZOWSKI_ALGEBRAIC )
            conv = new BrzozowskiAlgebraic( fsm );
        else
            conv = new StateElimination( fsm );

        conv->convert( ).toXML( out );
        delete conv;
    }
    else if( m_target == REGULAR_GRAMMAR )
    {
        if( m_algorithm == FA_LRG )
        {
            AbstractFAtoLRGConverter* conv = new FAtoLRGConverter( fsm );
            conv->convert( ).toXML( out );
            delete conv;
        }
        else
        {
            AbstractFAtoRRGConverter* conv = new FAtoRRGConverter( fsm );
            conv->convert( ).toXML( out );
            delete conv;
        }
    }
}

ConversionHandler::TFormalism ConversionHandler::parseFormalismFromString( const std::string & _target ) const
{
    string target;
    for( const auto & c : _target )
        target.append( 1, tolower( c ) );

    if( target == "fa" || target == "automaton" || target == "fsm" )
        return FINITE_AUTOMATA;
    if( target == "re" || target == "regexp" )
        return REGULAR_EXPRESSION;
    if( target == "rg" || target == "grammar" )
        return REGULAR_GRAMMAR;

    throw AlibException( "Unknown target type" );
}

ConversionHandler::TAlgorithm ConversionHandler::parseAlgorithmFromString( const std::string & _algorithm ) const
{
    string algorithm;
    for( const auto & c : _algorithm )
        algorithm.append( 1, tolower( c ) );

    if( algorithm == "" ) return DEFAULT;

    /* FA to RE */
    /* RG to RE */
    if( algorithm == "elimination" ) return STATE_ELIMINATION;
    if( algorithm == "algebraic" ) return BRZOZOWSKI_ALGEBRAIC;

    /* FA to RG */
    if( algorithm == "left" ) return FA_LRG;
    if( algorithm == "right" ) return FA_RRG;

    /* RE to FA */
    if( algorithm == "brzozowski" ) return BRZOZOWSKI_DERIVATION;
    if( algorithm == "glushkov" ) return GLUSHKOV;
    if( algorithm == "thompson" ) return THOMPSON;

    /* RE to RG */
    if( algorithm == "brzozowskirrg") return BRZOZOWSKI_RRG;

    /* RG to FA */
    if( algorithm == "rg" ) return RG_FA;

    throw AlibException( "Invalid conversion. See help." );
}

ConversionHandler::TAlgorithm ConversionHandler::getDefaultAlgorithm( void ) const
{
    if( m_source == FINITE_AUTOMATA && m_target == REGULAR_EXPRESSION ) return STATE_ELIMINATION;
    if( m_source == FINITE_AUTOMATA && m_target == REGULAR_GRAMMAR )    return FA_RRG;
    if( m_source == REGULAR_EXPRESSION && m_target == FINITE_AUTOMATA ) return GLUSHKOV;
    if( m_source == REGULAR_EXPRESSION && m_target == REGULAR_GRAMMAR ) return BRZOZOWSKI_RRG;
    if( m_source == REGULAR_GRAMMAR && m_target == FINITE_AUTOMATA )    return RG_FA;
    if( m_source == REGULAR_GRAMMAR && m_target == REGULAR_EXPRESSION ) return BRZOZOWSKI_ALGEBRAIC;

    throw AlibException( "ConversionHandler::getDefaultAlgorithm - unreachable code" );
}

bool ConversionHandler::checkAlgorithm( void ) const
{
    if( m_algorithm == DEFAULT ) return true;

    /* FA to RE */
    /* FA to RG */
    if( m_source == FINITE_AUTOMATA )
        return ( m_target == REGULAR_EXPRESSION &&
                    ( m_algorithm == BRZOZOWSKI_ALGEBRAIC || m_algorithm == STATE_ELIMINATION ) ) ||
               ( m_target == REGULAR_GRAMMAR &&
                    ( m_algorithm == FA_LRG || m_algorithm == FA_RRG ) );

    /* RE to FA */
    /* RE to RG */
    if( m_source == REGULAR_EXPRESSION )
        return ( m_target == FINITE_AUTOMATA &&
                    ( m_algorithm == BRZOZOWSKI_DERIVATION || m_algorithm == GLUSHKOV || m_algorithm == THOMPSON ) ) ||
               ( m_target == REGULAR_GRAMMAR &&
                    ( m_algorithm == BRZOZOWSKI_RRG) );

    /* RG to FA */
    /* RG to RE */
    if( m_source == REGULAR_GRAMMAR )
        return ( m_target == FINITE_AUTOMATA &&
                    ( m_algorithm == RG_FA ) ) ||
               ( m_target == REGULAR_EXPRESSION &&
                    ( m_algorithm == BRZOZOWSKI_ALGEBRAIC ) );

    return false;
}

} /* namespace conversions */
