/*
 * Glushkov.cpp
 *
 *  Created on: 11. 1. 2014
 *      Author: tomas
 */

#include "Glushkov.h"

using namespace alib;
using namespace alphabet;
using namespace automaton;
using namespace regexp;
using namespace std;

namespace conversions
{

Glushkov::Glushkov( const RegExp & re ) : AbstractREtoFAConverter( re )
{

}

Glushkov::~Glushkov( void )
{

}

FSM Glushkov::convert( void )
{
    // step 1
    initNumberSymbols( );

    for( const auto & symbol : RegExpAlphabet::getSymbols( m_re ) )
        m_fsm.addInputSymbol( symbol.getSymbol( ) );

    // steps 2, 3, 4
    constructBeginSymbolSet( );
    constructNeighbourSymbolSet( );
    constructEndSymbolSet( );
        // \e in q0 check is in step 7

    // step 5
    State q0( "q0" );
    m_fsm.addState( q0 );
    m_fsm.addInitialState( q0 );
    for( const auto & kv : m_numberedSymbols )
        m_fsm.addState( kv.second.m_state );

    // step 6
    for( const auto & ns : m_beginSymbolSet )
        m_fsm.addTransition( q0, ns.m_alphabetSymbol, ns.m_state );

    for( const auto & ns : m_neighbourSymbolSet )
    {
        const NumberedSymbol & first  = m_numberedSymbols.find( ns.m_first )->second;
        const NumberedSymbol & second = m_numberedSymbols.find( ns.m_second )->second;

        m_fsm.addTransition( first.m_state, second.m_alphabetSymbol, second.m_state );
    }

    // step 7
    for( const auto & ns : m_endSymbolSet )
        m_fsm.addFinalState( State ( ns.m_state ) );

    if( m_re.containsEmptyString( ) )
        m_fsm.addFinalState( q0 );

    return m_fsm;
}

void Glushkov::initNumberSymbols( void )
{
    int id = 1;

    for( const auto & symb : RegExpAlphabet::getSymbolsListInOrder( m_re ) )
        m_numberedSymbols.insert( pair<const RegExpSymbol*, NumberedSymbol>( symb, NumberedSymbol( symb, id ++ ) ) );
}

void Glushkov::constructBeginSymbolSet( void )
{
    for( const auto & s : getLeftmostSymbolsInTree( m_re.getRegExp( ) ) )
        m_beginSymbolSet.insert( m_numberedSymbols.find( s )->second );
}

void Glushkov::constructEndSymbolSet( void )
{
    for( const auto & s : getRightmostSymbolsInTree( m_re.getRegExp( ) ) )
        m_endSymbolSet.insert( m_numberedSymbols.find( s )->second );
}

void Glushkov::constructNeighbourSymbolSet( void )
{
    for( const auto & n : getNeighbours( m_re.getRegExp( ) ) )
        m_neighbourSymbolSet.insert( n );
}

// ----------------------------------------------------------------------------

set<const RegExpSymbol*> Glushkov::getLeftmostSymbolsInTree( const RegExpElement * node ) const
{
    const Alternation* alternation = dynamic_cast<const Alternation*>( node );
    const Concatenation* concatenation = dynamic_cast<const Concatenation*>( node );
    const Iteration* iteration = dynamic_cast<const Iteration*>( node );
    const RegExpSymbol* symbol = dynamic_cast<const RegExpSymbol*>( node );
    const RegExpEmpty* empty = dynamic_cast<const RegExpEmpty*>( node );
    const RegExpEpsilon* eps = dynamic_cast<const RegExpEpsilon*>( node );

    if( symbol )
        return getLeftmostSymbolsInTree( symbol );
    else if( alternation )
        return getLeftmostSymbolsInTree( alternation );
    else if( concatenation )
        return getLeftmostSymbolsInTree( concatenation );
    else if( iteration )
        return getLeftmostSymbolsInTree( iteration );
    else if( eps )
        return getLeftmostSymbolsInTree( eps );
    else if( empty )
        return getLeftmostSymbolsInTree( empty );

     throw AlibException( "Glushkov::getLeftmostSymbolsInTree - invalid RegExpElement node" );
}

set<const RegExpSymbol*> Glushkov::getLeftmostSymbolsInTree( const Alternation * node ) const
{
    set<const RegExpSymbol*> ret;

    for( const auto & e : node->getElements( ) )
    {
        const set<const RegExpSymbol*> tmp = getLeftmostSymbolsInTree( e );
        ret.insert( tmp.begin( ), tmp.end( ) );
    }

    return ret;
}

set<const RegExpSymbol*> Glushkov::getLeftmostSymbolsInTree( const Concatenation * node ) const
{
    set<const RegExpSymbol*> ret;

    for( const auto & e : node->getElements( ) )
    {
        set<const RegExpSymbol*> tmp = getLeftmostSymbolsInTree( e );
        ret.insert( tmp.begin( ), tmp.end( ) );

        if( ! e->containsEmptyString( ) ) // If this subtree can be epsilon, then we need to add next subtree also
            break;
    }

    return ret;
}

set<const RegExpSymbol*> Glushkov::getLeftmostSymbolsInTree( const Iteration * node ) const
{
    return getLeftmostSymbolsInTree( node->getElement( ) );
}

set<const RegExpSymbol*> Glushkov::getLeftmostSymbolsInTree( const RegExpSymbol * node ) const
{
    return set<const RegExpSymbol*> { node };
}

set<const RegExpSymbol*> Glushkov::getLeftmostSymbolsInTree( const RegExpEpsilon * node ) const
{
    return set<const RegExpSymbol*>( );
}

set<const RegExpSymbol*> Glushkov::getLeftmostSymbolsInTree( const RegExpEmpty * node ) const
{
    return set<const RegExpSymbol*>( );
}
// ----------------------------------------------------------------------------

set<const RegExpSymbol*> Glushkov::getRightmostSymbolsInTree( const RegExpElement * node ) const
{
    const Alternation* alternation = dynamic_cast<const Alternation*>( node );
    const Concatenation* concatenation = dynamic_cast<const Concatenation*>( node );
    const Iteration* iteration = dynamic_cast<const Iteration*>( node );
    const RegExpSymbol* symbol = dynamic_cast<const RegExpSymbol*>( node );
    const RegExpEmpty* empty = dynamic_cast<const RegExpEmpty*>( node );
    const RegExpEpsilon* eps = dynamic_cast<const RegExpEpsilon*>( node );

    if( symbol )
        return getRightmostSymbolsInTree( symbol );
    else if( alternation )
        return getRightmostSymbolsInTree( alternation );
    else if( concatenation )
        return getRightmostSymbolsInTree( concatenation );
    else if( iteration )
        return getRightmostSymbolsInTree( iteration );
    else if( eps )
        return getRightmostSymbolsInTree( eps );
    else if( empty )
        return getRightmostSymbolsInTree( empty );

     throw AlibException( "Glushkov::getRightmostSymbolsInTree - invalid RegExpElement node" );
}

set<const RegExpSymbol*> Glushkov::getRightmostSymbolsInTree( const Alternation * node ) const
{
    set<const RegExpSymbol*> ret;

    for( const auto & e : node->getElements( ) )
    {
        set<const RegExpSymbol*> tmp = getRightmostSymbolsInTree( e );
        ret.insert( tmp.begin( ), tmp.end( ) );
    }

    return ret;
}

set<const RegExpSymbol*> Glushkov::getRightmostSymbolsInTree( const Concatenation * node ) const
{
    set<const RegExpSymbol*> ret;

    for( auto it = node->getElements( ).rbegin( ); it != node->getElements( ).rend( ) ; it ++ )
    {
        set<const RegExpSymbol*> tmp = getRightmostSymbolsInTree( *it );
        ret.insert( tmp.begin( ), tmp.end( ) );

        if( ! ( * it )->containsEmptyString( ) )
            break;
    }

    return ret;
}

set<const RegExpSymbol*> Glushkov::getRightmostSymbolsInTree( const Iteration * node ) const
{
    return getRightmostSymbolsInTree( node->getElement( ) );
}

set<const RegExpSymbol*> Glushkov::getRightmostSymbolsInTree( const RegExpSymbol * node ) const
{
    return set<const RegExpSymbol*> { node };
}

set<const RegExpSymbol*> Glushkov::getRightmostSymbolsInTree( const RegExpEpsilon * node ) const
{
    return set<const RegExpSymbol*>( );
}

set<const RegExpSymbol*> Glushkov::getRightmostSymbolsInTree( const RegExpEmpty * node ) const
{
    return set<const RegExpSymbol*>( );
}

// ----------------------------------------------------------------------------

set<Glushkov::Neighbours> Glushkov::getNeighbours( const RegExpElement * node )  const
{
    const Alternation* alternation = dynamic_cast<const Alternation*>( node );
    const Concatenation* concatenation = dynamic_cast<const Concatenation*>( node );
    const Iteration* iteration = dynamic_cast<const Iteration*>( node );
    const RegExpSymbol* symbol = dynamic_cast<const RegExpSymbol*>( node );
    const RegExpEmpty* empty = dynamic_cast<const RegExpEmpty*>( node );
    const RegExpEpsilon* eps = dynamic_cast<const RegExpEpsilon*>( node );

    if( symbol )
        return getNeighbours( symbol );
    else if( alternation )
        return getNeighbours( alternation );
    else if( concatenation )
        return getNeighbours( concatenation );
    else if( iteration )
        return getNeighbours( iteration );
    else if( eps )
        return getNeighbours( eps );
    else if( empty )
        return getNeighbours( empty );

     throw AlibException( "Glushkov::getNeighbours - unknown RegExpElement* " );
}

set<Glushkov::Neighbours> Glushkov::getNeighbours( const Alternation * node ) const
{
    set<Neighbours> n;
    for( const auto & e : node->getElements( ) )
    {
        set<Neighbours> tmp = getNeighbours( e );
        n.insert( tmp.begin( ), tmp.end( ) );
    }

    return n;
}

set<Glushkov::Neighbours> Glushkov::getNeighbours( const Concatenation * node ) const
{
    set<Neighbours> n;
    for( const auto & e : node->getElements( ) )
    {
        set<Neighbours> tmp = getNeighbours( e );
        n.insert( tmp.begin( ), tmp.end( ) );
    }

    for( auto e = node->getElements( ).begin( ); e != node->getElements( ).end( ); e ++ )
    {
        auto f = e;
        if( f == node->getElements( ).end( ) )
            continue;

        for( f++ ; f != node->getElements( ).end( ); f ++ )
        {
            for( const auto & x : getRightmostSymbolsInTree( * e ) )
                for( const auto & y : getLeftmostSymbolsInTree( * f ) )
                    n.insert( Neighbours( x, y ) );

            if( ! ( * f )->containsEmptyString( ) )
                break;
        }
    }

    return n;
}

set<Glushkov::Neighbours> Glushkov::getNeighbours( const Iteration * node ) const
{
    set<Neighbours> n;
    set<Neighbours> tmp = getNeighbours( node->getElement( ) );
    n.insert( tmp.begin( ), tmp.end( ) );

    for( const auto & x : getRightmostSymbolsInTree( node->getElement( ) ) )
        for( const auto & y : getLeftmostSymbolsInTree( node->getElement( ) ) )
            n.insert( Neighbours( x, y ) );

    return n;
}

set<Glushkov::Neighbours> Glushkov::getNeighbours( const RegExpSymbol * node ) const
{
    return set<Neighbours>( );
}

set<Glushkov::Neighbours> Glushkov::getNeighbours( const RegExpEpsilon * node ) const
{
    return set<Neighbours>( );
}

set<Glushkov::Neighbours> Glushkov::getNeighbours( const RegExpEmpty * node ) const
{
    return set<Neighbours>( );
}

// ----------------------------------------------------------------------------

bool Glushkov::NumberedSymbol::operator<( const NumberedSymbol & x ) const
{
    return m_i < x.m_i;
}

Glushkov::NumberedSymbol::NumberedSymbol( const RegExpSymbol * symbol, int i ) :
        m_i( i ),
        m_alphabetSymbol( symbol->getSymbol( ) ),
        m_state( State( "[" + m_alphabetSymbol.getSymbol( ) + "], id " + to_string( m_i ) ) )
{

}

// ----------------------------------------------------------------------------

Glushkov::Neighbours::Neighbours( const RegExpSymbol * first, const RegExpSymbol * second ) :
        m_first( first ),
        m_second( second )
{

}

bool Glushkov::Neighbours::operator<( const Neighbours & x ) const
{
    if( m_first != x.m_first )
        return m_first < x.m_first;
    else
        return m_second < x.m_second;
}

} /* namespace conversions */
